from lingres.conll import read_conll09_file, write_conll09_file
from check_pd import extract_ud_coppola_constructions, extract_tiger_coppola_constructions

import sys

def swap_parent_child(tree, current_parent_node, current_child_node, new_relation, reattach_children = True, out_parent_label = None):
    current_child_node.remove_from_parent()
    if out_parent_label is None:
        out_parent_label = current_parent_node.incoming_edge_type
    if current_parent_node.parent is not None:
        current_parent_node.parent.add_child(current_child_node, out_parent_label)
    current_child_node.add_child(current_parent_node, new_relation)

    if reattach_children:
        for parent_child, label in list(current_parent_node.out_edges.items()):
            parent_child.remove_from_parent()
            current_child_node.add_child(parent_child, label)

def  node_tuple_set_to_indices(node_tuples):
    index_set = set()

    for node_tuple in node_tuples:
        index_set.add(tuple(map(lambda it: it.idx + 1, node_tuple)))

    return index_set

import itertools as it

def remove_pairs_with_same_verb(construction_pairs, tiger_indices, ud_indices):
    for pair in list(construction_pairs):
        for tiger_index in tiger_indices:
            for ud_index in ud_indices:
                if tiger_index[0] == ud_index[0]:
                    if not (tiger_index in construction_pairs and ud_index in construction_pairs):
                        continue
                    if tiger_index in construction_pairs:
                        construction_pairs.remove(tiger_index)
                        construction_pairs.remove(ud_index)

class Transformation:
    def __init__(self):
        self.miss_count = 0
        self.wrong_hit_count = 0

    def transform_tiger_sentences(self, sentences):
        for sentence in sentences:
            self.transform_tiger_to_ud(sentence.dependency_tree)
            yield sentence

    def count_numbers_of_correct_tiger_ud_transforms(self, tiger_sentences, ud_sentences):
        for sent_idx, tiger_sent, ud_sent in zip(it.count(0), tiger_sentences, ud_sentences):
            tiger_indices = node_tuple_set_to_indices(self._find_tiger_constructions(tiger_sent.dependency_tree))
            ud_indices =  node_tuple_set_to_indices(self._find_ud_constructions(ud_sent.dependency_tree))

            incorrect_construction_pairs = tiger_indices.symmetric_difference(ud_indices)

            remove_pairs_with_same_verb(incorrect_construction_pairs, tiger_indices, ud_indices)

            if len(incorrect_construction_pairs) > 0:
                print("TiGer constructions not in UD", tiger_sent.original_sentence_id)
                print("Tiger not UD construction")
                print(tiger_indices - ud_indices)
                print("UD constructions not in TiGer")
                print(ud_indices - tiger_indices)
                print()

            self.miss_count += len(ud_indices - tiger_indices)
            self.wrong_hit_count += len(tiger_indices - ud_indices)

            self.transform_tiger_to_ud(tiger_sent.dependency_tree)
            yield tiger_sent

    def transform_ud_sentences(self, sentences):
        for sentence in sentences:
            self.transform_ud_to_tiger(sentence.dependency_tree)
            yield sentence

    def transform_tiger_to_ud(self, tree):
        pass

    def transform_ud_to_tiger(self, tree):
        pass

    def _find_tiger_constructions(self, tree):
        pass

    def _find_ud_constructions(self, tree):
        pass

    def _find_tiger_construction(self, tree):
        pass

class CoordTransformation(Transformation):
    def _find_ud_constructions(self, tree):
        constructions = set()
        for node in tree.iter_token_nodes_left_to_right():
            coord_nodes = []
            if node.incoming_edge_type is not None \
                    and node.incoming_edge_type in ("conj", "cc"):
                continue
            for child in sorted(node.children, key = lambda c: c.token.idx):
                if child.incoming_edge_type in ("conj", "cc"):
                    coord_nodes.append(child)

            if len(coord_nodes) > 0:
                constructions.add((node,) + tuple(coord_nodes))

        return constructions

    def transform_ud_to_tiger(self, tree):
        ud_construction = self._find_ud_constructions(tree)

        for construction in ud_construction:
            prev_node = construction[0]
            for node in construction[1:]:
                prev_node.add_child(node, node.incoming_edge_type)
                #node.parent = prev_node
                prev_node = node

    def transform_tiger_to_ud(self, tree):
        tiger_constructions = self._find_tiger_constructions(tree)

        for construction in tiger_constructions:
            root_node = construction[0]
            for node in construction[1:]:
                root_node.add_child(node, node.incoming_edge_type)

    def _find_tiger_constructions(self, tree):
        constructions = set()

        for node in tree.iter_token_nodes_left_to_right():
            if node.incoming_edge_type is not None \
                    and node.incoming_edge_type in ("conj", "cc"):
                        continue

            coord_nodes = []
            curr_node = node
            while True:
                for child in curr_node.children:
                    if child.incoming_edge_type in ("conj", "cc"):
                        coord_nodes.append(child)
                        curr_node = child
                        break
                else:
                    break

            if len(coord_nodes) > 0:
                constructions.add((node,) + tuple(coord_nodes))

        return constructions


class PrepTransformation(Transformation):
    def _find_tiger_constructions(self, tree):
        constructions = set()
        for node in tree.nodes:
            if node.incoming_edge_type == "NK":
                #and node.parent.token.pos in ("APPR", "APPRART", "APPO"):
                constructions.add((node.parent, node))

        return constructions

    def transform_tiger_to_ud(self, tree):
        tiger_constructions = self._find_tiger_constructions(tree)
        for prep_node, child_node in tiger_constructions:
            swap_parent_child(tree, prep_node, child_node, "case", reattach_children = False)

    def _find_ud_constructions(self, tree):
        constructions = set()
        for node in tree.nodes:
            if node.incoming_edge_type == "case":
                constructions.add((node, node.parent))

        return constructions

    def transform_ud_to_tiger(self, tree):
        ud_construction = self._find_ud_constructions(tree)
        for prep_node, child_node in ud_construction:
            swap_parent_child(tree, child_node, prep_node, "NK", reattach_children = False)

class PDCopTransformation(Transformation):

    def transform_tiger_sentences(self, sentences):
        for sentence in sentences:
            self.transform_tiger_to_ud(sentence.dependency_tree)
            yield sentence

    def count_numbers_of_correct_tiger_ud_transforms(self, tiger_sentences, ud_sentences):
        for sent_idx, tiger_sent, ud_sent in zip(it.count(0), tiger_sentences, ud_sentences):
            tiger_indices = node_tuple_set_to_indices(self._find_tiger_constructions(tiger_sent.dependency_tree))
            ud_indices =  node_tuple_set_to_indices(self._find_ud_constructions(ud_sent.dependency_tree))

            incorrect_construction_pairs = tiger_indices.symmetric_difference(ud_indices)

            remove_pairs_with_same_verb(incorrect_construction_pairs, tiger_indices, ud_indices)

            if len(incorrect_construction_pairs) > 0:
                print("Incorrect Pairs in Sentence", tiger_sent.original_sentence_id)
                print("PD-NK not cop construction")
                print(tiger_indices - ud_indices)
                print("cop not PD-NK")
                print(ud_indices - tiger_indices)
                print()

            self.miss_count += len(ud_indices - tiger_indices)
            self.wrong_hit_count += len(tiger_indices - ud_indices)

            self.transform_tiger_to_ud(tiger_sent.dependency_tree)
            yield tiger_sent

    def transform_ud_to_tiger(self, tree):
        ud_construction = self._find_ud_constructions(tree)

        for parent_node, verb_node in ud_construction:
            swap_parent_child(tree, verb_node, parent_node, "PD")

    def transform_tiger_to_ud(self, tree):
        tiger_constructions = self._find_tiger_constructions(tree)

        for parent_node, verb_node in tiger_constructions:
            swap_parent_child(tree, parent_node, verb_node, "cop")

    def _find_tiger_constructions(self, tree):
        constructions = set()
        for node in tree.nodes:
            if node.incoming_edge_type == "PD" and not (node.token.morph_dict.get("VerbForm") == "Part" and (node.parent.token.lemma == "haben")):
                constructions.add((node.parent, node))

        return constructions

    def _find_ud_constructions(self, tree):
        constructions = set()
        for node in tree.nodes:
            if node.incoming_edge_type == "cop":
                constructions.add((node, node.parent))

        return constructions

def transform_trees(ud_sents, tiger_sents):
    for ud_sent, tiger_sent in zip(ud_sents, tiger_sents):
        ud_coppola = extract_ud_coppola_constructions(ud_sent)
        tiger_coppola = extract_tiger_coppola_constructions(tiger_sent)

        for pd_constr in tiger_coppola:
            if pd_constr in ud_coppola:
                ud_coppola.discard(pd_constr)
            else:
                for ud_constr in ud_coppola:
                    if ud_constr[0] == pd_constr[1]:
                        ud_coppola.discard(ud_constr)
                        break
                else:
                    continue

            pd_node = tiger_sent.dependency_tree.nodes[pd_constr[1]]
            pd_parent_node = tiger_sent.dependency_tree.nodes[pd_constr[0]]

            pd_node.remove_from_parent()
            pd_node.add_child(pd_parent_node, "cop")

            pd_parent_edges = list(pd_parent_node.out_edges.items())

            ud_cop_parent_node = ud_sent.dependency_tree.nodes[pd_constr[1]]
            ud_cop_edge_children_indices = set(map(lambda n: n.idx, ud_cop_parent_node.children))
            for child, label in pd_parent_edges:
                if child.idx in ud_cop_edge_children_indices:
                    pd_node.add_child(child, label)

        yield tiger_sent

import argparse

if __name__ == "__main__":
    transform_class_name = sys.argv[1]
    transform_direction_description = sys.argv[2]
    from_sents = read_conll09_file(sys.argv[3])
#    to_sents = read_conll09_file(sys.argv[4])

    transform_cls = globals()[transform_class_name]
    transform = transform_cls()

    if transform_direction_description == "tig-ud":
        converted_sents = transform.transform_tiger_sentences(from_sents)
    elif transform_direction_description == "ud-tig":
        converted_sents = transform.transform_ud_sentences(from_sents)
    else:
        raise RuntimeError("Unknown transform direction {}".format(transform_direction_description))
    write_conll09_file(sys.argv[4], converted_sents)
#    print("Misses:", transform.miss_count)
#    print("Wrong Hits:", transform.wrong_hit_count)

    sys.exit(0)

    ud_sents = read_conll09_file(sys.argv[1])
    tiger_sents = read_conll09_file(sys.argv[2])

    converted_tiger_trees = transform_trees(ud_sents, tiger_sents)

    write_conll09_file(sys.argv[3], converted_tiger_trees)

