# -*- coding: utf8 -*-

import itertools

from .tabular import TabularReader, TabularWriter


def read_clearnlp_file(fname, skip_errors = False):
    reader = TabularReader(fname)
    token_builder = TabularTokenBuilder(form_idx = 1, lemma_idx = None, pos_idx = 3, morph_idx = None, original_id_idx = 0)
    decoder = TabularDecoder(token_builder, [TabularDependencyTreeBuilder(id_idx = 0, parent_id_idx = 6, edge_type_idx = 7, no_parent_id = "0")])

    for idx, sent in enumerate(reader.read_sentences()):
        try:
            yield decoder.decode_sentence(sent)
        except Exception as e:
            if skip_errors:
                print ("Sentence {} is not coded correctly".format(idx + 1))
            else:
                raise RuntimeError("Sentence {} is not coded correctly".format(idx + 1)) from e

def read_conll09_file(filename, skip_errors = False):
    reader = TabularReader(filename)
    token_builder = TabularTokenBuilder(form_idx = 1, lemma_idx = 2, pos_idx = 3, morph_idx = 5, original_id_idx = 0)
    decoder = TabularDecoder(token_builder, [TabularDependencyTreeBuilder(id_idx = 0, parent_id_idx = 6, edge_type_idx = 7, no_parent_id = "0")])

    for idx, sent in enumerate(reader.read_sentences()):
        try:
            yield decoder.decode_sentence(sent)
        except Exception as e:
            if skip_errors:
                print ("Sentence {} is not coded correctly".format(idx + 1))
            else:
                raise RuntimeError("Sentence {} is not coded correctly".format(idx + 1)) from e

def write_conll09_file(filename, iterator):
    writer = TabularWriter(filename)
    encoder = TabularEncoder([SingleIDEncoder(0), TabularTokenEncoder(form_idx = 1, lemma_idx = 2, pos_idx = 3, morph_idx = 5), TabularDepedencyTreeEncoder(parent_id_idx = 6, edge_type_idx = 7, no_parent_id = "0")])
    for sent in iterator:
        enc_sent = encoder.encode_sentence(sent)
        writer.write_sentence(enc_sent)

class DependencyTree:
    def __init__(self):
        self.nodes = []
        self.roots = []

    def append_token(self, token, parent = None, edge_type = None, **args):
        node = DependencyTreeNode(token, self, len(self.nodes), **args)
        self.nodes.append(node)
        if parent:
            parent.add_child(node, edge_type)
        else:
            self.roots.append(node)

        return node

    def __str__(self):
        return "\n".join(map(lambda n: str(n), self.roots))

    def __iter__(self):
        return iter(self.nodes)

    def node_for_token(self, token):
        for node in self.nodes:
            if node.token == token:
                return node
        else:
            return None

    def iter_token_nodes_left_to_right(self):
        for node in self.nodes:
            yield node

class Sentence:
    def __init__(self, tokens, dependency_tree = None, original_sentence_id = None):
        self.tokens = tokens
        self.dependency_tree = dependency_tree
        self.original_sentence_id = original_sentence_id

    def __iter__(self):
        return iter(self.tokens)

    def __len__(self):
        return len(self.tokens)

    def __getitem__(self, idx):
        return self.tokens[idx]

    @property
    def raw_str(self):
        return " ".join(map(lambda t: t.form, self))

class Token:
    def __init__(self, form = None, lemma = None, pos = None, coarse_pos = None, morph = None, original_id = None, morph_dict = None):
        self.form = form
        self.lemma = lemma
        self.pos = pos
        self.coarse_pos = coarse_pos
        self.morph = morph
        self.original_id = original_id

        self.morph_dict = morph_dict

def decode_morph(morph):
    if not morph:
        return {}

    result = {}
    for entry in morph.split("|"):
        res = entry.rsplit("=", 1)
        if len(res) > 1:
            key, val = res
            result[key] = val

    return result


class TabularTokenBuilder:
    def __init__(self, form_idx, lemma_idx, pos_idx, original_id_idx = None, coarse_pos_tag_idx = None, morph_idx = None):
        self.form_idx = form_idx
        self.lemma_idx = lemma_idx
        self.pos_idx = pos_idx
        self.coarse_pos_tag_idx = coarse_pos_tag_idx
        self.morph_idx = morph_idx
        self.original_id_idx = original_id_idx

    def decode_line(self, line):
        if self.coarse_pos_tag_idx:
            coarse_pos_tag = line[self.coarse_pos_tag_idx]
        else:
            coarse_pos_tag = None

        if self.morph_idx:
            morph = line[self.morph_idx]
        else:
            morph = None

        morph_dict = decode_morph(morph)

        if self.original_id_idx is not None:
            orig_id = line[self.original_id_idx]
        else:
            orig_id = None

        if self.lemma_idx is not None:
            lemma = line[self.lemma_idx]
        else:
            lemma = None

        token = Token(form = line[self.form_idx], lemma = lemma, pos = line[self.pos_idx], coarse_pos = coarse_pos_tag, morph = morph, original_id = orig_id, morph_dict = morph_dict)
        return token

class DependencyTreeNode:
    def __init__(self, token, tree, idx):
        self.token = token
        self.out_edges = {}
        self.tree = tree
        self.parent = None
        self.incoming_edge_type = None
        self.idx = idx

    @property
    def children(self):
        return self.out_edges.keys()

    def add_child(self, child, edge_type):
        if child.parent == None:
            self.tree.roots.remove(child)
        else:
            child.parent.remove_child(child)
        self.out_edges[child] = edge_type
        child.parent = self
        child.incoming_edge_type = edge_type

    def remove_child(self, child, root_edge_label = "--"):
        assert child.parent == self
        assert child in self.children

        del self.out_edges[child]
        child.parent = None
        child.incoming_edge_type = root_edge_label
        self.tree.roots.append(child)

    def remove_from_parent(self):
        if self.parent is not None:
            self.parent.remove_child(self)

    def line_repr(self, indent = 0):
        if indent == 0:
            indent_str = ""
        elif indent == 1:
            indent_str = "+----"
        else:
            indent_str = "     " * (indent - 1) + "+----"
        own_line = indent_str + "({}) {}".format(self.incoming_edge_type, self.token)
        lines = [(self.idx, own_line)]
        for child in self.children:
            lines += child.line_repr(indent + 1)

        return lines

    def __str__(self, indent = 0):
        return "\n".join(map(lambda t: t[1], sorted(self.line_repr())))

class TabularDependencyTreeBuilder:
    def __init__(self, id_idx, parent_id_idx, edge_type_idx, no_parent_id = "0", use_idx_as_id = False):
        self.id_idx = id_idx
        self.parent_id_idx = parent_id_idx
        self.edge_type_idx = edge_type_idx
        self.no_parent_id = no_parent_id
        self.use_idx_as_id = use_idx_as_id

    def extract_id_from_id_str(self, id_str):
        split_res = id_str.split("_")
        if len(split_res) == 2:
            return split_res[1]
        else:
            return id_str

    def decode_sentence_lines(self, sentence, lines):
        tree = DependencyTree()
        node_idx_map = {}
        idx = 1
        for line, token in zip(lines, sentence):
            node = tree.append_token(token)

            if self.use_idx_as_id:
                tok_id = str(idx)
            else:
                tok_id = self.extract_id_from_id_str(line[self.id_idx])
            node_idx_map[tok_id] = node

            idx += 1

        for idx, line_and_node in enumerate(zip(lines, tree.nodes)):
            line, node = line_and_node
            parent_id = line[self.parent_id_idx]
            if parent_id != self.no_parent_id:
                try:
                    parent = node_idx_map[parent_id]
                except KeyError:
                    raise RuntimeError("Parent node {} of node {} could not be found".format(parent_id, line[self.id_idx]))
                parent.add_child(node, line[self.edge_type_idx])

        sentence.dependency_tree = tree

class TabularDecoder:
    def __init__(self, token_level_decoder, sentence_level_decoders):
        self.token_level_decoder = token_level_decoder
        self.sentence_level_decoders = sentence_level_decoders

    def decode_document(self, doc):
        sents = []
        for sent in doc:
            sents.append(self.decode_sentence(sent))

        return sents

    def decode_sentence(self, sentence):
        decoded_tokens = []
        for idx, tok in enumerate(sentence):
            new_tok = self.token_level_decoder.decode_line(tok)
            decoded_tokens.append(new_tok)
            new_tok.idx = idx

        decoded_sent = Sentence(decoded_tokens)
        if decoded_tokens[0].original_id is not None:
            id_split = decoded_tokens[0].original_id.split("_")

            if len(id_split) == 2:
                decoded_sent.original_sentence_id = id_split[0]

        for tok in decoded_tokens:
            tok.sentence = decoded_sent

        for sent_lvl_decoder in self.sentence_level_decoders:
            sent_lvl_decoder.decode_sentence_lines(decoded_sent, sentence)

        return decoded_sent

class TabularEncoder:
    def __init__(self, encoders):
        self.encoders = encoders

    def encode_sentence(self, sent):
        encoded_sent = []
        sent_encodings = [list(map(lambda e: e.items(), enc.encode_sentence(sent))) for enc in self.encoders]
        for tok_encs in zip(*sent_encodings):
            token_cols = sorted(itertools.chain(*tok_encs))
            token = []
            for tok_idx, col_val in token_cols:
                while tok_idx > len(token):
                    token.append("--")
                if col_val == None:
                    col_val = "--"
                token.append(col_val)
            encoded_sent.append(token)

        return encoded_sent

class TabularDepedencyTreeEncoder:
    def __init__(self, parent_id_idx, edge_type_idx, no_parent_id):
        self.parent_id_idx = parent_id_idx
        self.edge_type_idx = edge_type_idx
        self.no_parent_id = no_parent_id

    def encode_sentence(self, sent):
        encoded_sent = []

        tree = sent.dependency_tree
        for idx, node in enumerate(tree.iter_token_nodes_left_to_right()):
            encoded_tok = {}
            if node.parent:
                encoded_tok[self.parent_id_idx] = str(node.parent.idx + 1)
            else:
                encoded_tok[self.parent_id_idx] = self.no_parent_id
            encoded_tok[self.edge_type_idx] = node.incoming_edge_type
            encoded_sent.append(encoded_tok)

        return encoded_sent

class SingleIDEncoder:
    def __init__(self, id_idx):
        self.id_idx = id_idx

    def encode_sentence(self, sentence):
        return list(map(lambda x: dict([(self.id_idx, str(x))]), range(1, len(sentence) + 1)))

class TabularTokenEncoder:
    def __init__(self, form_idx, lemma_idx, pos_idx, coarse_pos_tag_idx = None, morph_idx = None):
        self.form_idx = form_idx
        self.lemma_idx = lemma_idx
        self.pos_idx = pos_idx
        self.coarse_pos_tag_idx = coarse_pos_tag_idx
        self.morph_idx = morph_idx

    def encode_sentence(self, sentence):
        encoded_tokens = []
        for tok in sentence:
            encoded_tokens.append(self.encode_token(tok))
        return encoded_tokens

    def encode_token(self, tok):
        encoded_dict = {}
        encoded_dict[self.form_idx] = tok.form
        encoded_dict[self.lemma_idx] = tok.lemma
        encoded_dict[self.pos_idx] = tok.pos
        if self.coarse_pos_tag_idx:
            encoded_dict[self.coarse_pos_tag_idx] = tok.coarse_pos

        if self.morph_idx:
            encoded_dict[self.morph_idx] = tok.morph

        return encoded_dict

